﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using EZ_B;
using EZ_Builder;

namespace Dynamixel {

  public partial class FormMain : EZ_Builder.UCForms.FormPluginMaster {

    public enum VersionEnum {
      Protocol_1,
      Protocol_2
    }

    public enum PortTypeEnum {
      UART,
      DigitalPort
    }

    DynamixelV1 _v1 = new DynamixelV1();
    DynamixelV2 _v2 = new DynamixelV2();

    // This is a duplicate of the _cf.CustomObjectv2 for performance
    ConfigServos _servoConfig = new ConfigServos();

    public FormMain() {

      InitializeComponent();
    }

    void displayVersion() {

      try {

        string xmlFilename = Common.CombinePath(Constants.PLUGINS_FOLDER, _cf._pluginGUID, "plugin.xml");

        EZ_Builder.Config.Sub.PluginV1XML xml = (EZ_Builder.Config.Sub.PluginV1XML)Common.DeserializeObjectFile(xmlFilename, typeof(EZ_Builder.Config.Sub.PluginV1XML));

        Invokers.SetAppendText(tbLog, true, "Version: {0}", xml.VersionRelease);
      } catch {

        Invokers.SetAppendText(tbLog, true, "Unable to read plugin.xml");
      }
    }

    private void Form1_Load(object sender, EventArgs e) {

      displayVersion();

      // Bind to the events for moving a servo and changing connection state
      EZBManager.EZBs[0].OnConnectionChange += FormMain_OnConnectionChange;
      EZBManager.EZBs[0].Servo.OnServoMove += Servo_OnServoMove;
      EZBManager.EZBs[0].Servo.OnServoRelease += Servo_OnServoRelease;
      EZBManager.EZBs[0].Servo.OnServoSpeed += Servo_OnServoSpeed;

      Invokers.SetAppendText(tbLog, true, "Connected Events");

      if (EZBManager.EZBs[0].IsConnected) {

        initUART();

        disableAllFeedback();
      }
    }

    void FormMain_OnConnectionChange(bool isConnected) {

      // If the connection is established, send an initialization to the ez-b for the uart which we will be using
      if (isConnected) {

        if (_cf.STORAGE[ConfigTitles.GLOBAL_SERVO_POSITIONS] != null)
          EZ_B.Servo.SERVO_MAX = Convert.ToInt32(_cf.STORAGE[ConfigTitles.GLOBAL_SERVO_POSITIONS]);

        initUART();

        disableAllFeedback();
      }
    }

    private void FormMain_FormClosing(object sender, FormClosingEventArgs e) {

      EZBManager.EZBs[0].Servo.OnServoMove -= Servo_OnServoMove;
      EZBManager.EZBs[0].Servo.OnServoRelease -= Servo_OnServoRelease;
      EZBManager.EZBs[0].Servo.OnServoSpeed -= Servo_OnServoSpeed;
    }

    private void initUART() {

      if (((PortTypeEnum)_cf.STORAGE[ConfigTitles.PORT_TYPE]) == PortTypeEnum.DigitalPort)
        return;

      if (EZBManager.EZBs[0].IsConnected) {

        UInt32 baud = Convert.ToUInt32(_cf.STORAGE[ConfigTitles.BAUD_RATE]);

        Invokers.SetAppendText(tbLog, true, "Init {0}bps", baud);

        EZBManager.EZBs[0].Uart.UARTExpansionInit(1, baud);
      }
    }

    public override void SetConfiguration(EZ_Builder.Config.Sub.PluginV1 cf) {

      cf.STORAGE.AddIfNotExist(ConfigTitles.BAUD_RATE, 1000000);

      cf.STORAGE.AddIfNotExist(ConfigTitles.GLOBAL_SERVO_POSITIONS, EZ_B.Servo.SERVO_MAX);

      cf.STORAGE.AddIfNotExist(ConfigTitles.PORT_TYPE, PortTypeEnum.UART);

      cf.STORAGE.AddIfNotExist(ConfigTitles.DIGITAL_PORT, Digital.DigitalPortEnum.D0);

      EZ_B.Servo.SERVO_MAX = Convert.ToInt32(cf.STORAGE[ConfigTitles.GLOBAL_SERVO_POSITIONS]);

      // If the servo config is empty, assign a blank one
      if (cf._customObjectEncodedV2.Length == 0)
        cf.SetCustomObjectV2(_servoConfig);

      _servoConfig = (ConfigServos)cf.GetCustomObjectV2(typeof(ConfigServos));

      base.SetConfiguration(cf);
    }

    public override EZ_Builder.Config.Sub.PluginV1 GetConfiguration() {

      _cf.SetCustomObjectV2(_servoConfig);

      return base.GetConfiguration();
    }

    void serialWrite(byte[] data) {

      if (((PortTypeEnum)_cf.STORAGE[ConfigTitles.PORT_TYPE]) == PortTypeEnum.DigitalPort) {

        var baudRate = (Uart.BAUD_RATE_ENUM)Enum.Parse(typeof(Uart.BAUD_RATE_ENUM), "Baud_" + _cf.STORAGE[ConfigTitles.BAUD_RATE].ToString());

        EZBManager.EZBs[0].Uart.SendSerial((Digital.DigitalPortEnum)_cf.STORAGE[ConfigTitles.DIGITAL_PORT], baudRate, data);
      } else {

        EZBManager.EZBs[0].Uart.UARTExpansionWrite(1, data);
      }
    }

    public override void SendCommand(string windowCommand, params string[] values) {

      if (windowCommand.Equals("setled", StringComparison.InvariantCultureIgnoreCase)) {

        if (values.Length != 2)
          throw new Exception(string.Format("Expecting 2 parameters, you passed {1}", values.Length));

        Servo.ServoPortEnum servoPort = new EZ_Builder.Scripting.HelperPortParser(values[0]).ServoPort;

        if (servoPort < EZ_B.Servo.ServoPortEnum.V0 || servoPort > EZ_B.Servo.ServoPortEnum.V99)
          throw new Exception("Only virtual servos are supported for dynamixel. Virtual servos start with a 'v', such as v1, v2, v3..");

        bool status = EZ_Builder.Scripting.Helper.GetTrueOrFalse(values[1]) == EZ_Builder.Scripting.Helper.TrueFalseEnum.True;

        var servoConfig = _servoConfig.GetPort(servoPort);

        if (servoConfig == null)
          throw new Exception(string.Format("Virtual Servo {0} is not configured for dynamixel usage", servoPort));

        if (servoConfig.Version == VersionEnum.Protocol_1)
          serialWrite(_v1.LED(servoPort, status));
        else
          serialWrite(_v2.LED(servoPort, status));
      } else {

        base.SendCommand(windowCommand, values);
      }
    }

    public override object[] GetSupportedControlCommands() {

      List<string> cmds = new List<string>();

      cmds.Add("SetLED, Port, true|false");

      return cmds.ToArray();
    }

    void Servo_OnServoRelease(Servo.ServoPortEnum[] servos) {

      List<byte> cmdData = new List<byte>();

      foreach (var servoPort in servos) {

        if (servoPort < EZ_B.Servo.ServoPortEnum.V0 || servoPort > EZ_B.Servo.ServoPortEnum.V99)
          continue;

        var servoConfig = _servoConfig.GetPort(servoPort);

        if (servoConfig == null)
          continue;

        byte id = (byte)(servoPort - EZ_B.Servo.ServoPortEnum.V0);

        if (servoConfig.Version == VersionEnum.Protocol_1)
          cmdData.AddRange(_v1.ReleaseServo(id));
        else
          cmdData.AddRange(_v2.ReleaseServo(id));
      }

      if (cmdData.Count != 0)
        serialWrite(cmdData.ToArray());
    }

    void Servo_OnServoSpeed(EZ_B.Classes.ServoSpeedItem[] servos) {

      List<byte> cmdData = new List<byte>();

      foreach (var servo in servos) {

        if (servo.Port < EZ_B.Servo.ServoPortEnum.V0 || servo.Port > EZ_B.Servo.ServoPortEnum.V99)
          continue;

        var servoConfig = _servoConfig.GetPort(servo.Port);

        if (servoConfig == null)
          continue;

        byte id = (byte)(servo.Port - EZ_B.Servo.ServoPortEnum.V0);
        int speed = servo.Speed * 51;

        if (servoConfig.Version == VersionEnum.Protocol_1)
          cmdData.AddRange(_v1.ServoSpeed(id, speed));
        else
          cmdData.AddRange(_v2.ServoSpeed(id, speed));
      }

      if (cmdData.Count != 0)
        serialWrite(cmdData.ToArray());
    }

    void Servo_OnServoMove(EZ_B.Classes.ServoPositionItem[] servos) {

      List<byte> cmdData = new List<byte>();

      foreach (var servo in servos) {

        if (servo.Port < EZ_B.Servo.ServoPortEnum.V0 || servo.Port > EZ_B.Servo.ServoPortEnum.V99)
          continue;

        var servoConfig = _servoConfig.GetPort(servo.Port);

        if (servoConfig == null)
          continue;

        int position = (int)EZ_B.Functions.RemapScalar(servo.Position, Servo.SERVO_MIN, Servo.SERVO_MAX, 0, servoConfig.MaxPosition - 1);

        byte id = (byte)(servo.Port - EZ_B.Servo.ServoPortEnum.V0);

        if (servoConfig.Version == VersionEnum.Protocol_1)
          cmdData.AddRange(_v1.MoveServoCmd(id, position));
        else
          cmdData.AddRange(_v2.MoveServoCmd(id, position));
      }

      if (cmdData.Count != 0)
        serialWrite(cmdData.ToArray());
    }

    private void ucConfigurationButton1_Click(object sender, EventArgs e) {

      using (FormDynamixelConfig form = new FormDynamixelConfig()) {

        form.SetConfiguration(_cf);

        if (form.ShowDialog() == System.Windows.Forms.DialogResult.OK) {

          _cf = form.GetConfiguration();

          initUART();

          _servoConfig = (ConfigServos)_cf.GetCustomObjectV2(typeof(ConfigServos));

          disableAllFeedback();

          EZ_B.Servo.SERVO_MAX = Convert.ToInt32(_cf.STORAGE[ConfigTitles.GLOBAL_SERVO_POSITIONS]);
        }
      }
    }

    void disableAllFeedback() {

      if (!EZBManager.EZBs[0].IsConnected)
        return;

      Invokers.SetAppendText(tbLog, true, "Disable Status Packet");

      List<byte> cmdData = new List<byte>();

      foreach (var servo in _servoConfig.Servos)
        if (servo.Version == VersionEnum.Protocol_1) {

          cmdData.AddRange(_v1.DisableStatusPacket(servo.Port));
          cmdData.AddRange(_v1.DisableAllAlarms(servo.Port));
        } else {

          cmdData.AddRange(_v2.DisableStatusPacket(servo.Port));
          cmdData.AddRange(_v2.DisableAllAlarms(servo.Port));
        }

      serialWrite(cmdData.ToArray());
    }

    private void btnForceInit_Click(object sender, EventArgs e) {

      if (!EZBManager.EZBs[0].IsConnected) {

        MessageBox.Show("You must be connected to an EZ-B to send the initialization.");

        return;
      }

      try {

        initUART();

        disableAllFeedback();
      } catch (Exception ex) {

        MessageBox.Show("Dynamixel Error: " + ex.Message);
      }
    }
  }
}
